# CreateApiGenerator.py - Creates an object representing a create API
import os
from jinja2 import Environment, Template, FileSystemLoader
from smoacks.sconfig import sconfig

class SqlAlchemyGenerator:
    def __init__(self, app_object):
        self._app_object = app_object
        self.name = self._app_object.name

    def getField(self, prop):
        if prop.type == 'string':
            if prop.isId:
                fk_text = ", ForeignKey('{}.{}')".format(prop.foreignKey, prop.name) if prop.foreignKey else ""
                return "{} = Column(BINARY(16){}, primary_key=True)".format(prop.name, fk_text)
            elif prop.format == 'date':
                return "{} = Column(DateTime)".format(prop.name)
            elif prop.format == 'uuid':
                fk_text = ", ForeignKey('{}.{}')".format(prop.foreignKey, prop.name) if prop.foreignKey else ""
                return "{} = Column(BINARY(16){})".format(prop.name, fk_text)
            elif prop.maxLength and prop.maxLength > 0:
                return "{} = Column(String({}))".format(prop.name, prop.maxLength)
            else:
                return "{} = Column(String(80))".format(prop.name)
        elif prop.type == 'number':
            if prop.format == 'double':
                return "{} = Column(Double)".format(prop.name)
            else:
                return "{} = Column(Float)".format(prop.name)
        elif prop.type == 'integer':
            if prop.format == 'int64':
                return "{} = Column(Long)".format(prop.name)
            else:
                return "{} = Column(Integer)".format(prop.name)
        elif prop.type == 'object':
            return "{} = Column(JSON)".format(prop.name)
        elif prop.type == 'boolean':
            return "{} = Column(Boolean)".format(prop.name)
        else:
            raise ValueError("Property {} has invalid type {}".format(prop.name, prop.type))

    def getJinjaDict(self):
        # Establish constant values and the overall dictionary structure
        result = {
            'name': self.name,
            'snakeName': self._app_object.getSnakeName(),
            'mixedName': self.name,
            'dmFields': [],
            'genprefix': sconfig['structure']['genprefix'],
            'gensubdir': sconfig['structure']['gensubdir'],
            'idCount': self._app_object._idCount,
            'relationships': [],
            'uuid_set': set()
        }
        # Loop through the properties and update the structure where needed
        properties = self._app_object.getAllProperties()
        for prop in properties:
            if prop.isId:
                result['name_id'] = prop.name
                result['dmFields'].append(self.getField(prop))
                result['uuid_set'].add(prop.name)
            if not prop.isId and not (prop.name in ['record_created', 'record_updated']):
                result['dmFields'].append(self.getField(prop))
                if prop.format == 'uuid':
                    result['uuid_set'].add(prop.name)
        # Loop through relationships
        for rel in self._app_object.relationships:
            result['relationships'].append({
              'name': rel,
              'table': self._app_object.relationships[rel]['table'],
              'field': self._app_object.relationships[rel]['field']
            })
        return result

    def render(self):
        env = Environment(
            loader = FileSystemLoader('templates')
        )
        template = env.get_template('SQLAlchemyModel.jinja')
        gendir = os.path.join(sconfig['structure']['root'],
                              sconfig['structure']['datamodeldir'],
                              sconfig['structure']['gensubdir'])
        if not os.path.isdir(gendir):
            os.makedirs(gendir, exist_ok=True)
        module_filename = os.path.join(gendir, "__init__.py")
        if not os.path.isfile(module_filename):
            initfile = open(module_filename, "w")
            initfile.close()
        outfile = open(os.path.join(gendir, "{}{}.py".format(sconfig['structure']['genprefix'], self.name)), "w")
        outfile.write(template.render(self.getJinjaDict()))
        outfile.close()
        filedir = os.path.join(sconfig['structure']['root'],
                               sconfig['structure']['datamodeldir'])
        module_filename2 = os.path.join(gendir, "__init__.py")
        if not os.path.isfile(module_filename2):
            initfile2 = open(module_filename2, "w")
            initfile2.close()
        # We should not overwrite customization file if it exists
        dmo_filename = os.path.join(filedir, "{}.py".format(self.name))
        if not os.path.isfile(dmo_filename):
            template2 = env.get_template('DataModelObject.jinja')
            of2 = open(dmo_filename, "w")
            of2.write(template2.render(self.getJinjaDict()))
            of2.close()



# structure.py - Routines for generation app structure
import os
import stat
from jinja2 import Environment, Template, TemplateError, UndefinedError, TemplateNotFound, FileSystemLoader
from smoacks.sconfig import sconfig

class SmoacksStructure:
    def __init__(self):
        self.name = 'SmoacksStructure'
        self.env = [
            {'template': 'api_util.jinja',
             'dir': sconfig['structure']['utildir'],
             'outfile': "api_util.py",
             'module_dir': True},
            {'template': 'app-env.jinja',
             'dir': sconfig['structure']['bindir'],
             'outfile': "app-env"},
            {'template': 'base.jinja',
             'dir': os.path.join(sconfig['structure']['datamodeldir'],
                                 sconfig['structure']['gensubdir']),
             'outfile': 'base.py',
             'module_dir': True},
            {'template': 'dev-api-server.jinja',
             'dir': sconfig['structure']['kubedir'],
             'outfile': "dev-api-server.yaml"},
            {'template': 'Dockerfile.jinja',
             'outfile': "Dockerfile"},
            {'template': 'gitignore.jinja',
             'outfile': ".gitignore",
             'overwrite': False},
            {'template': 'local-env.jinja',
             'dir': sconfig['structure']['bindir'],
             'outfile': "local-env"},
            {'template': 'schema.jinja',
             'dir': sconfig['structure']['specdir'],
             'outfile': sconfig['parameters']['source_spec'],
             'overwrite': False},
            {'template': 'server-loop.jinja',
             'dir': sconfig['structure']['sourcedir'],
             'outfile': "server-loop",
             'executable': True},
            {'template': 'server_logging.jinja',
             'dir': sconfig['structure']['sourcedir'],
             'outfile': "server_logging.yaml"},
            {'template': 'server.jinja',
             'dir': sconfig['structure']['sourcedir'],
             'outfile': "server.py"},
            {'template': 'shutdown.jinja',
             'dir': sconfig['structure']['apiobjectdir'],
             'outfile': "shutdown.py",
             'module_dir': True},
            {'template': 'test-api-server.jinja',
             'dir': sconfig['structure']['kubedir'],
             'outfile': "test-api-server.yaml"},
            {'template': 'testme.jinja',
             'dir': sconfig['structure']['bindir'],
             'outfile': 'testme',
             'executable': True},
            {'template': 'TestUtil.jinja',
             'dir': sconfig['structure']['testdir'],
             'outfile': 'TestUtil.py',
             'module_dir': True}
        ]
        self.template_dict = sconfig['env_defaults']
#        self.template_dict['smoacks_local_dev_path'] = os.getcwd()
        self.template_dict['smoacks_local_dev_path'] = os.path.join(sconfig['structure']['pvPathRoot'],
                                                                    sconfig['env_defaults']['smoacks_app_name'])

    def renderEnvironment(self):
        env = Environment(
            loader = FileSystemLoader('templates')
        )
        # Ensure there is a coverage directory with a runType.bash in it
        if not os.path.isdir('coverage'):
              coverpath = os.path.join(sconfig['structure']['root'], 'coverage')
              os.makedirs(coverpath, exist_ok=True)
              rtfile = open(os.path.join(coverpath, 'runType.bash'), "w")
              rtfile.write('export run_coverage=0')
              rtfile.close()
        for filespec in self.env:
           template = env.get_template(filespec['template'])
           filedir = os.path.join(sconfig['structure']['root'], filespec['dir']) if 'dir' in filespec and filespec['dir'] else sconfig['structure']['root']
           if not os.path.isdir(filedir):
              os.makedirs(filedir, exist_ok=True)
              if 'module_dir' in filespec and filespec['module_dir']:
                  tmf_file = open(os.path.join(filedir, '__init__.py'), "w")
                  tmf_file.close()
           outfilename = os.path.join(filedir, filespec['outfile'])
           if not os.path.isfile(outfilename) or 'overwrite' not in filespec or filespec['overwrite']:
               outfile = open(outfilename, "w")
               try:
                   filestring = template.render(self.template_dict)
                   outfile.write(filestring)
                   if 'executable' in filespec and filespec['executable']:
                       perm = os.stat(os.path.join(filedir, filespec['outfile']))
                       os.chmod(os.path.join(filedir, filespec['outfile']), perm.st_mode | stat.S_IEXEC)
                   outfile.close() 
               except TemplateNotFound:
                   print('Caught a TemplateNotFoundError on {}'.format(filespec['template']))
               except UndefinedError:
                   print('Caught a UndefinedError on {}'.format(filespec['template']))
               except TemplateError:
                   print('Caught a TemplateError on {}'.format(filespec['template']))
               except:
                   print('Caught an untyped error on {}'.format(filespec['template']))
           else:
               print('Skipping {} because it exists and should not be overwritten'.format(filespec['outfile']))



import copy
import os
import logging
import sys
import yaml
from openapi_spec_validator import validate_spec
from smoacks.sconfig import sconfig
from smoacks.Schema import Schema, scr_schemas
from smoacks.AppObject import AppObject, scr_objects
from smoacks.ConnexionApiGenerator import ConnexionApiGenerator
from smoacks.NoseTestGenerator import NoseTestGenerator
from smoacks.OpenapiGenerator import OpenapiGenerator
from smoacks.ServerGenerator import ServerGenerator
from smoacks.SqlAlchemyGenerator import SqlAlchemyGenerator

LOGGER = logging.getLogger('ApiGenerator')


def generate_code():
    SPEC_FILENAME = os.path.join(sconfig['structure']['root'],
                                 sconfig['structure']['specdir'],
                                 sconfig['parameters']['source_spec'])
    LOGGER.info('Generating application from {}'.format(SPEC_FILENAME))

    # Load the configuration yaml file
    try:
        OPENAPI_SPEC = yaml.load(open(SPEC_FILENAME), Loader=yaml.FullLoader)
    except:
        print("Error opening / parsing {}, likely missing, unreadable, nor not YAML".format(SPEC_FILENAME))
        exit(2)

    VALIDATE_COPY = copy.deepcopy(OPENAPI_SPEC)

    # Validate that the provided YAML is an OpenAPI specification
    # For details see https://pypi.org/project/openapi-spec-validator/#description
    try:
        validate_spec(VALIDATE_COPY)
    except:
        print("Specification {} is not a valid OpenAPI 3.0 specification".format(SPEC_FILENAME))
        exit(3)

    # Find where the schema definitions exist in the API spec
    SCHEMAS_YAML = OPENAPI_SPEC['components']['schemas']

    # TODO: Better job handling update vs. identity; better
    #       handling of missing descriptions
    for schema in SCHEMAS_YAML:
        print('Schema: {}'.format(schema))
        # Create the global dictionary of schemas
        scr_schemas[schema] = Schema(schema, SCHEMAS_YAML[schema])
        # If this schema object identifies an AppObject, then create it
        # (if needed), and set the identity object property on it
        objectDesc = scr_schemas[schema].description
        if scr_schemas[schema].identityObject:
            print("--> Identity Object!")
            objectName = scr_schemas[schema].identityObject
            if objectName not in scr_objects:
                scr_objects[objectName] = AppObject(objectName, objectDesc)
            scr_objects[objectName].addSchema(scr_schemas[schema])
        # If this schema object updates an AppObject, then create it
        # (if needed), and set the update object property on it
        if scr_schemas[schema].updateObject:
            print("--> Update Object!")
            objectName = scr_schemas[schema].updateObject
            if objectName not in scr_objects:
                scr_objects[objectName] = AppObject(objectName, objectDesc)
            scr_objects[objectName].addSchema(scr_schemas[schema])
        if scr_schemas[schema].extendedObject:
            print("--> Extended Object!")
            objectName = scr_schemas[schema].extendedObject
            if objectName not in scr_objects:
                scr_objects[objectName] = AppObject(objectName, objectDesc)
            scr_objects[objectName].addSchema(scr_schemas[schema])

    # Now that we know all of the app objects are created, walk through the
    # schemas again to see if there are any to use for parameters or responses
    # for specific verbs
    for schemaName in scr_schemas:
        schema = scr_schemas[schemaName]
        if schema.paramVerb:
            app_object = scr_objects[schema.appObject]
            app_object.addParamVerb(schema)
        if schema.respVerb:
            app_object = scr_objects[schema.appObject]
            app_object.addRespVerb(schema)

    # If the parameters say to not include login, remove it from the path output
    if not sconfig['parameters']['include_shutdown']:
        print('Removing /shutdown path')
        OPENAPI_SPEC['paths'].pop('/shutdown')

    # Create the schema object that will hold the definition for the /login endpoint
    LOGIN_SCHEMA = {
        'schema': {
            'type': 'object',
            'properties': {}
        }
    }
    # Generate the schemas
    for object_name in scr_objects:
        LOGIN_SCHEMA['schema']['properties'][object_name + 's'] = {
            'type': 'array',
            'items': {
                '$ref': '#/components/schemas/' + object_name
            }
        }
        print('Handling {}'.format(object_name))
        sqlalchemy_generator = SqlAlchemyGenerator(scr_objects[object_name])
        sqlalchemy_generator.render()
        connexion_generator = ConnexionApiGenerator(scr_objects[object_name])
        connexion_generator.render()
        nosetest_generator = NoseTestGenerator(scr_objects[object_name])
        nosetest_generator.render()
        openapi_generator = OpenapiGenerator(scr_objects[object_name])
        api_obj = openapi_generator.render_to_yaml_obj()
        OPENAPI_SPEC['paths'].update(api_obj)

    # Build login get path, used to hydrate user data at login
    OPENAPI_SPEC['paths']['/login'] = {
        'get': {
            'summary': 'Hydrate application',
            'tags': [sconfig['env_defaults']['smoacks_app_name']],
            'description': 'Returns initial login data for this service',
            'responses': {
                '200': {
                    'description': 'Successful response',
                    'content': {
                        'application/json': LOGIN_SCHEMA
                    }
                }
            }
        }
    }

    server_generator = ServerGenerator(scr_objects)
    server_generator.render()
    filedir = os.path.join(sconfig['structure']['root'], sconfig['structure']['specdir'])
    if not os.path.isdir(filedir):
        os.makedirs(filedir, exist_ok=True)
    outfile = open(os.path.join(filedir, sconfig['parameters']['dest_spec']), "w")
    yaml.dump(OPENAPI_SPEC, outfile, default_flow_style=False)



# Schema.py - Tools for understanding OpenAPI 3.0 schema definitions
from smoacks.Property import Property

scr_schemas = dict()

# This handles a subset of the OpenAPI schema specification.
# It is intended to include the core components of the specification
# that are suitable for direct persistence into a SQL-based database
# Specifically, it can handle a top-level typed object that is defined
# with a property  list, or a property list combined with another
# schema reference via allOf
class Schema:

    def __init__(self, name, schemaYaml):
        self.name = name
        self._yaml = schemaYaml
        self._properties = dict()
        self._references = []
        self.description = self._yaml['description'] if 'description' in self._yaml else None
        self.identityObject = self._yaml['x-smoacks-create'] if 'x-smoacks-create' in self._yaml else None
        self.extendedObject = self._yaml['x-smoacks-extended'] if 'x-smoacks-extended' in self._yaml else None
        self.updateObject = self._yaml['x-smoacks-update'] if 'x-smoacks-update' in self._yaml else None
        self.appObject = self._yaml['x-smoacks-object'] if 'x-smoacks-object' in self._yaml else None
        self.paramVerb = self._yaml['x-smoacks-api-verb-param'] if 'x-smoacks-api-verb-param' in self._yaml else None
        self.respVerb = self._yaml['x-smoacks-api-verb-resp'] if 'x-smoacks-api-verb-resp' in self._yaml else None
        self.emitTestData = self._yaml['x-smoacks-test-data'] if 'x-smoacks-test-data' in self._yaml else True
        self.relationships = self._yaml['x-smoacks-relationships'] if 'x-smoacks-relationships' in self._yaml else None
        propertiesYaml = None
        if 'properties' in self._yaml:
            propertiesYaml = self._yaml['properties']
        elif 'allOf' in self._yaml:
            allOf = self._yaml['allOf']
            for item in allOf:
                if 'type' in item and item['type'] == 'object':
                    propertiesYaml = item['properties']
                elif '$ref' in item:
                    self._references.append(item['$ref'])
        if propertiesYaml is not None:
            for propertyName in propertiesYaml:
                self._properties[propertyName] = Property(self.name, propertyName, propertiesYaml[propertyName])

    # Allow contents to be printed when casting to string
    def __str__(self):
        return str(self.__dict__)

    def getProperties(self):
        result = self._properties.copy()
        for ref in self._references:
            pieces = ref.split('/')
            ref_schema_name = pieces[-1]
            print('Found ref_schema_name: {}'.format(ref_schema_name))
            result.update(scr_schemas[ref_schema_name].getProperties())
        return result

    def getProperty(self, propertyName):
        return self._properties[propertyName]



# util.py - Utility functions that are used in multople places
import re

# This code came from a StackOverflow answer
# https://stackoverflow.com/questions/19053707/converting-snake-case-to-lower-camel-case-lowercamelcase/19053800
def to_camelcase(value):
    components = value.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + ''.join(x.title() for x in components[1:])

def to_mixedcase(value):
    components = value.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return ''.join(x.title() for x in components)

# This code came from a StackOverflow answer
# https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
def to_snakecase(value):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', value)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


# AppObject.py - Application objects
from smoacks.util import to_camelcase, to_mixedcase, to_snakecase

scr_objects = dict()

class AppObject:
    def __init__(self, name, desc = None):
        self.name = name
        self.description = desc if desc else name
        self._schemas = []
        self._properties = []
        self._paramVerbs = {}
        self._respVerbs = {}
        self._idCount = 0
        self._idProperty = None
        self.identitySchemaName = None
        self.extendedSchemaName = None
        self.emitTestData = False
        self.relationships = None
    
    def addSchema(self, schema):
        self._schemas.append(schema)
        if schema.identityObject and not self.identitySchemaName:
            self.identitySchemaName = schema.name
            if schema.relationships:
                self.relationships = schema.relationships
            for prop in schema.getProperties().values():
                self._properties.append(prop)
                if prop.isId:
                    self._idProperty = prop
                    self._idCount += 1
        if schema.extendedObject:
            self.extendedSchemaName = schema.name

    def addParamVerb(self, schema):
        self._paramVerbs[schema.paramVerb] = schema.name

    def addRespVerb(self, schema):
        self._respVerbs[schema.respVerb] = schema.name

    def getSnakeName(self):
        return to_snakecase(self.name)
    
    def getCamelName(self):
        return to_camelcase(self.name)
    
    def getMixedName(self):
        return to_mixedcase(self.name)

    def getCreateObject(self):
        # Loop through the properties and update the structure where needed
        result = {}
        for prop in self._properties:
            if (not prop.isId or self._idCount > 1) and not prop.readOnly:
                result[prop.name] = prop.example
        return result

    def getAllProperties(self):
        return self._properties



# NoseTestGenerator.py - Creates nose tests for a model API
import os
from jinja2 import Environment, Template, FileSystemLoader
from smoacks.sconfig import sconfig
from smoacks.AppObject import scr_objects

class NoseTestGenerator:
    def __init__(self, app_object):
        self._app_object = app_object
        self.name = self._app_object.name

    def getJinjaDict(self):
        # Establish constant values and the overall dictionary structure
        result = {
            'name': self.name,
            'hasSearch': False,
            'idCount': self._app_object._idCount,
            'idsString': None,
            'snakeName': self._app_object.getSnakeName(),
            'createObj': self._app_object.getCreateObject(),
            'foreignKeys': []
        }
        properties = self._app_object.getAllProperties()
        getAsserts = []
        unitTestEditObject = {}
        unitTestAssert = None
        for prop in properties:
            if prop.searchField:
                result['hasSearch'] = True
            if prop.foreignKey:
                fk_app_object = scr_objects[prop.foreignKey]
                result['foreignKeys'].append({
                    'name': prop.name,
                    'createObj': fk_app_object.getCreateObject(),
                    'snakeName': fk_app_object.getSnakeName(),
                    'idField': fk_app_object._idProperty.name
                })
            if prop.isId:
                result['name_id'] = prop.name
                if not result['idsString']:
                    result['idsString'] = 'added_{}'.format(prop.name) if prop.foreignKey else "'{}'".format(prop.example)
                else:
                    result['idsString'] += ','
                    result['idsString'] += 'added_{}'.format(prop.name) if prop.foreignKey else "'{}'".format(prop.example)
            # We need to change a value in unit tests of PUT verb
            elif prop.example != None and not prop.readOnly:
                if prop.editUnitTest:
                    unitTestEditObject[prop.name] = prop.editUnitTest
                    unitTestAssert = 'assert json["{}"] == {}'.format(prop.name, prop.getUnitTestLiteral())
                    getAsserts.append('assert json["{}"] == {}'.format(prop.name, prop.getExamplePythonLiteral()))
                else:
                    unitTestEditObject[prop.name] = prop.example
                    if prop.foreignKey:
                        getAsserts.append('assert json["{}"] == added_{}'.format(prop.name, prop.name))
                    else:
                        getAsserts.append('assert json["{}"] == {}'.format(prop.name, prop.getExamplePythonLiteral()))
        result['getAsserts'] = getAsserts
        result['unitTestEditObject'] = str(unitTestEditObject)
        result['unitTestAssert'] = unitTestAssert
        return result

    def render(self):
        env = Environment(
            loader = FileSystemLoader('templates')
        )
        template = env.get_template('NoseTests.jinja')
        filedir = os.path.join(sconfig['structure']['root'], sconfig['structure']['testdir'])
        if not os.path.isdir(filedir):
            os.makedirs(filedir, exist_ok=True)
        outfile = open(os.path.join(filedir, "test-{}-api.py".format(self._app_object.getSnakeName())), "w")
        outfile.write(template.render(self.getJinjaDict()))



# ConnexionApiGenerator.py - Creates an object implementing a Connexion API
import os
from jinja2 import Environment, Template, FileSystemLoader
from smoacks.sconfig import sconfig

class ConnexionApiGenerator:
    def __init__(self, app_object):
        self._app_object = app_object
        self.name = self._app_object.name

    def getJinjaDict(self):
        # Establish constant values and the overall dictionary structure
        result = {
            'name': self.name,
            'idList': None,
            'pkList': None,
            'idCount': self._app_object._idCount,
            'hasSearch': False
        }
        # Loop through the properties and update the structure where needed
        properties = self._app_object.getAllProperties()
        for prop in properties:
            if prop.isId:
                result['name_id'] = prop.name
                if not result['idList']:
                    result['idList'] = prop.name
                    result['pkList'] = self.name + '.' + prop.name
                else:
                    result['idList'] += ', ' + prop.name
                    result['pkList'] += ', ' + self.name + '.' + prop.name
            if prop.searchField:
                result['hasSearch'] = True
                result['search_field'] = prop.name
        result.update(sconfig['env_defaults'])
        return result

    def render(self):
        env = Environment(
            loader = FileSystemLoader('templates')
        )
        template = env.get_template('ConnexionAPIs.jinja')
        filedir = os.path.join(sconfig['structure']['root'], sconfig['structure']['apiobjectdir'])
        if not os.path.isdir(filedir):
            os.makedirs(filedir, exist_ok=True)
        outfilename = os.path.join(filedir, "{}s.py".format(self._app_object.getSnakeName()))
        if not os.path.isfile(outfilename):
            outfile = open(outfilename, "w")
            outfile.write(template.render(self.getJinjaDict()))
            outfile.close()



# sconfig.py - smoacks configuration file
import os
import yaml
import logging

structure_overrides = ['SMOACKS_ROOT']
parameters_overrides = ['SMOACKS_SOURCE_SPEC', 'SMOACKS_DEST_SPEC',
                        'SMOACKS_SERVER_CONTAINER_PORT', 'SMOACKS_INCLUDE_LOGIN']
env_defaults_overrides = ['SMOACKS_APP_NAME']

LOGGER = logging.getLogger('SMOACKS')

sconfig = None
custom_config = None

with open('conf/smoacks_default.yaml', 'r') as yaml_def_config:
    sconfig = yaml.load(yaml_def_config, Loader=yaml.FullLoader)

if os.path.isfile('conf/smoacks.yaml'):
    with open('conf/smoacks.yaml', 'r') as yaml_config:
        custom_config = yaml.load(yaml_config, Loader=yaml.FullLoader)
        if 'structure' in custom_config:
            sconfig['structure'].update(custom_config['structure'])
        if 'env_defaults' in custom_config:
            sconfig['env_defaults'].update(custom_config['env_defaults'])
        if 'parameters' in custom_config:
            sconfig['parameters'].update(custom_config['parameters'])

sconfig['env_defaults']['smoacks_app_name_c'] = sconfig['env_defaults']['smoacks_app_name'].upper()
sconfig['env_defaults']['smoacks_app_name_k'] = sconfig['env_defaults']['smoacks_app_name'].replace('_', '-')
# Promote spec name from parameters into env defaults so it can more easily
# be used in jinja templates during generation
sconfig['env_defaults']['dest_spec'] = sconfig['parameters']['dest_spec']

for ovr in structure_overrides:
    if ovr in os.environ:
        sconfig['structure'][ovr[8:].lower()] = os.environ[ovr]
for ovr in parameters_overrides:
    if ovr in os.environ:
        sconfig['parameters'][ovr[8:].lower()] = os.environ[ovr]
for ovr in env_defaults_overrides:
    if ovr in os.environ:
        sconfig['env_defaults'][ovr.lower()] = os.environ[ovr]

LOGGER.debug('SMOACKS configuration: {}'.format(str(sconfig)))



from smoacks.sconfig import sconfig
from smoacks.structure import SmoacksStructure
from smoacks.ApiGenerator import generate_code

def main():
    my_struct = SmoacksStructure()
    print ('Setting up SMOACKS for this project with bindir: {}'.format(sconfig['structure']['bindir']))
    my_struct.renderEnvironment()

def gen():
    generate_code()



# ServerGenerator.py - Creates an object implementing a Connexion API
import os
from jinja2 import Environment, Template, FileSystemLoader
from smoacks.sconfig import sconfig

class ServerGenerator:
    def __init__(self, app_objects):
        self._app_objects = app_objects

    def get_import(self, app_object):
        return "from dm.{} import {}".format(app_object.name, app_object.name)

    def getJinjaDict(self):
        # Establish constant values and the overall dictionary structure
        result = {
            'dmImports': [],
            'objects': [],
            'gensubdir': sconfig['structure']['gensubdir']
        }
        # Loop through the properties and update the structure where needed
        set_first_table = False
        for objName in self._app_objects:
            result['dmImports'].append(self.get_import(self._app_objects[objName]))
            if not set_first_table:
                set_first_table = True
                result['first_table_name'] = objName
            result['objects'].append({
                'name': objName,
                'snake_name': self._app_objects[objName].getSnakeName() + 's',
                'table_name': objName
            })
        return result

    def render(self):
        env = Environment(
            loader = FileSystemLoader('templates')
        )
        template = env.get_template('DataModel.jinja')
        filedir = os.path.join(sconfig['structure']['root'], sconfig['structure']['datamodeldir'])
        tmf_file_name = os.path.join(filedir, '__init__.py') 
        if not os.path.isfile(tmf_file_name):
            tmf_file = open(tmf_file_name, "w")
            tmf_file.close()
        outfile = open(os.path.join(filedir, 'DataModel.py'), "w")
        outfile.write(template.render(self.getJinjaDict()))
        template2 = env.get_template('login.jinja')
        filedir2 = os.path.join(sconfig['structure']['root'], sconfig['structure']['apiobjectdir'])
        outfile2 = open(os.path.join(filedir2, 'login.py'), "w")
        outfile2.write(template2.render(self.getJinjaDict()))
        template3 = env.get_template('test-login-api.jinja')
        filedir3 = os.path.join(sconfig['structure']['root'], sconfig['structure']['testdir'])
        outfile3 = open(os.path.join(filedir3, 'test-login-api.py'), "w")
        outfile3.write(template3.render(self.getJinjaDict()))



from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

setup(name='smoacks',
      version='0.2.6',
      description='Simple Microservices with OpenAPI, Connexion, Kubernetes, and SQLAlchemy',
      url='https://github.com/wittlesouth/smoacks',
      author='Wittle South Ventures, LLC',
      author_email='service@wittlesouth.com',
      license='MIT',
      packages=['smoacks'],
      include_package_data=True,
      data_files=[('conf', ['conf/smoacks_default.yaml']),
                  ('templates', ['templates/api_util.jinja',
                                 'templates/app-env.jinja',
                                 'templates/base.jinja',
                                 'templates/ConnexionApis.jinja',
                                 'templates/dev-api-server.jinja',
                                 'templates/DataModel.jinja',
                                 'templates/DataModelObject.jinja',
                                 'templates/Dockerfile.jinja',
                                 'templates/gitignore.jinja',
                                 'templates/local-env.jinja',
                                 'templates/login.jinja',
                                 'templates/ModelApis.jinja',
                                 'templates/NoseTests.jinja',
                                 'templates/schema.jinja',
                                 'templates/server_logging.jinja',
                                 'templates/server-loop.jinja',
                                 'templates/server.jinja',
                                 'templates/shutdown.jinja',
                                 'templates/SQLAlchemyModel.jinja',
                                 'templates/test-api-server.jinja',
                                 'templates/test-login-api.jinja',
                                 'templates/testme.jinja',
                                 'templates/TestUtil.jinja'])],
      install_requires=[
          'connexion',
          'coverage',
          'flask_jwt_extended',
          'jinja2',
          'nose',
          'openapi-spec-validator',
          'PyYAML',
          'SQLAlchemy'
          ],
      entry_points={
          'console_scripts': ['smoacks-setup=smoacks.command_line:main',
                              'smoacks-gen=smoacks.command_line:gen']
      },
      long_description=long_description,
      long_description_content_type='text/markdown',
      zip_safe=False)



# Property.py - Handles OpenAPI 3.0 properties, with extended smoacks attributes
import json
from smoacks.util import to_camelcase, to_mixedcase, to_snakecase

# This handles a subset of the OpenAPI properties specification.
# It is intended to include the core components of the specification
# that are suitable for direct persistence into a SQL-based database
class Property:
    def __init__(self, schemaName, name, propertyYaml):
        self.schemaName = schemaName
        self.name = name
        self._yaml = propertyYaml
        if 'type' in propertyYaml:
            self.type = propertyYaml['type']
            self.baseObject = propertyYaml['x-smoacks-base-object'] if 'x-smoacks-base-object' in propertyYaml else False
            self.createOnly = propertyYaml['x-smoacks-create-only'] if 'x-smoacks-create-only' in propertyYaml else False
            self.description = propertyYaml['description'] if 'description' in propertyYaml else None
            self.editUnitTest = propertyYaml['x-smoacks-edit-unit-test'] if 'x-smoacks-edit-unit-test' in propertyYaml else None
            self.enum = propertyYaml['enum'] if 'enum' in propertyYaml else None
            self.example = propertyYaml['example'] if 'example' in propertyYaml else None
            self.exclusiveMaximum = propertyYaml['exclusiveMaximum'] if 'exclusiveMaximum' in propertyYaml else False
            self.exclusiveMinimum = propertyYaml['exclusiveMinimum'] if 'exclusiveMinimum' in propertyYaml else False
            self.foreignKey = propertyYaml['x-smoacks-foreign-key'] if 'x-smoacks-foreign-key' in propertyYaml else None
            self.format = propertyYaml['format'] if 'format' in propertyYaml else None
            self.isId = propertyYaml['x-smoacks-model-id'] if 'x-smoacks-model-id' in propertyYaml else False
            self.maximum = propertyYaml['maximum'] if 'maximum' in propertyYaml else None
            self.maxLength = propertyYaml['maxLength'] if 'maxLength' in propertyYaml else None
            self.minimum = propertyYaml['minimum'] if 'minimum' in propertyYaml else None
            self.minLength = propertyYaml['minLength'] if 'minLength' in propertyYaml else None
            self.nullable = propertyYaml['nullable'] if 'nullable' in propertyYaml else False
            self.pattern = propertyYaml['pattern'] if 'pattern' in propertyYaml else None
            self.searchField = propertyYaml['x-smoacks-search-field'] if 'x-smoacks-search-field' in propertyYaml else False
            self.stringFormat = propertyYaml['format'] if 'format' in propertyYaml else None
            self.readOnly = propertyYaml['readOnly'] if 'readOnly' in propertyYaml else False
        elif '$ref' in propertyYaml:
            self.ref = propertyYaml['$ref']

    # ID is special cased in Redux Objects
    def getCamelName(self):
        if self.name != to_snakecase(self.schemaName) + "_id":
            return to_camelcase(self.name)
        else:
            return "id"

    # ID is special cased in Redux Objects
    def getMixedName(self):
        if self.name != to_snakecase(self.schemaName) + "_id":
            return to_mixedcase(self.name)
        else:
            return "Id"

    # ID is special cased in Redux Objects
    def getUpperName(self):
        if self.name != to_snakecase(self.schemaName) + "_id":
            return self.name.upper()
        else:
            return "ID"

    # Returns a string for what the default value of this property should be
    # assuming use only in JavaScript (Redux Object)
    def defaultValue(self):
        if self.type == 'string':
            if self.format == 'date':
                return "new Date(0)"
            else:
                return "''"
        elif self.type == 'date':
            return "new Date(0)"
        elif self.type == 'object':
            return "Map({})"
        else:
            return "0"

    # Gets example value in form suitable for JavaScript
    def getExampleJs(self):
        if self.type == 'string' and self.example:
            return "'" + self.example + "'"
        elif self.type == 'object' and self.example:
            return json.dumps(self.example)
        elif not self.example:
            return 'undefined'
        else:
            return self.example

    # Gets example value in form suitable for Python
    def getExamplePythonLiteral(self):
        if self.type == 'string' and self.example:
            return "'" + self.example + "'"
        elif self.type == 'object' and self.example:
            return json.dumps(self.example)
        elif self.example == None:
            return 'None'
        else:
            return self.example

    # Gets example value in form suitable for JavaScript
    def getUnitTestLiteral(self):
        if self.type == 'string' and self.editUnitTest:
            return "'" + self.editUnitTest + "'"
        elif self.type == 'object' and self.editUnitTest:
            return json.dumps(self.editUnitTest)
        elif not self.editUnitTest:
            return 'None'
        else:
            return self.editUnitTest

    # Gets invalid value for validation tests
    def getInvalidValue(self):
        if self.type == 'string':
            if self.minLength:
                return "'X'"
            elif self.maxLength:
                return "'" + "x" * (self.maxLength+1) + "'"
            elif not self.nullable:
                return 'undefined'
            else:
                return 10
        elif self.type == 'object':
            return 'undefined'
        else:
            return 'undefined'

    # Gets true if this property should have clent validations
    def needsValidation(self):
        result = False
        if (self.readOnly):
            # Any readOnly property requires no client validation
            result = False
        elif self.maxLength or self.minLength or self.minimum or self.maximum or self.enum:
            result = True
        elif not result and not self.nullable:
            result = True
        elif self.pattern:
            result = True
        return result

    def needsInputTransform(self):
        result = False
        if (self.type == 'string' and self.format == 'date'):
            result = True
        return result

    def needsOutputTransform(self):
        result = False
        if (self.type == 'string' and self.format == 'date'):
            result = True
        if (self.type == 'object'):
            result = True
        return result

    # Allow contents to be printed when casting to string
    def __str__(self):
        return str(self.__dict__)





# CreateApiGenerator.py - Creates an object representing a create API
import os
from jinja2 import Environment, Template, FileSystemLoader
import yaml
from smoacks.sconfig import sconfig
from smoacks.Schema import scr_schemas

class OpenapiGenerator:
    def __init__(self, app_object):
        self._app_object = app_object
        self.name = self._app_object.name

    def getJinjaDict(self):
        # Establish constant values and the overall dictionary structure
        ao = self._app_object
        result = {
            'name': self.name,
            'idList': [],
            'snakeName': self._app_object.getSnakeName(),
            'mixedName': self.name,
            'identitySchemaName': self._app_object.identitySchemaName,
            'extendedSchemaName': self._app_object.extendedSchemaName,
            'hasSearch': False,
            'paramVerbSchema': {
                'post': ao._paramVerbs['post'] if 'post' in ao._paramVerbs else ao.identitySchemaName,
                'search': 'schemas/{}'.format(ao._paramVerbs['search']) if 'search' in ao._paramVerbs else 'parameters/search_text',
                'getVerb': ao._paramVerbs['get'] if 'get' in ao._paramVerbs else ao.identitySchemaName,
                'put': ao._paramVerbs['put'] if 'put' in ao._paramVerbs else ao.identitySchemaName
            },
            'respVerbSchema': {
                'search': ao._respVerbs['search'] if 'search' in ao._respVerbs else ao.identitySchemaName,
                'getVerb': ao._respVerbs['get'] if 'get' in ao._respVerbs else ao.identitySchemaName
            },
        }
        result.update(sconfig['env_defaults'])
        # Loop through the properties and update the structure where needed
        properties = self._app_object.getAllProperties()
        for prop in properties:
            if prop.isId:
                result['name_id'] = prop.name
                result['idList'].append(prop.name)
            if prop.searchField:
                result['hasSearch'] = True
        return result

    def render_to_yaml_obj(self):
        env = Environment(
            loader = FileSystemLoader('templates')
        )
        template = env.get_template('ModelAPIs.jinja')
        rendered_string = template.render(self.getJinjaDict())
#        print('rendered_string = ', rendered_string)
        result = yaml.load(rendered_string, Loader=yaml.FullLoader)
        return result
